from io import BytesIO
from io import SEEK_END

from clikit.utils._compat import encode

from .stream_input_stream import StreamInputStream


class StringInputStream(StreamInputStream):
    """
    An input stream that reads from a string.
    """

    def __init__(self, string=""):  # type: (str) -> None
        self._stream = BytesIO()

        super(StringInputStream, self).__init__(self._stream)

        self.set(string)

    def clear(self):  # type: () -> None
        self._stream.truncate(0)
        self._stream.seek(0)

    def set(self, string):  # type: (str) -> None
        self.clear()

        self._stream.write(encode(string))
        self._stream.seek(0)

    def append(self, string):  # type: (str) -> None
        pos = self._stream.tell()
        self._stream.seek(0, SEEK_END)
        self._stream.write(encode(string))
        self._stream.seek(pos)



from typing import List
from typing import Optional

from clikit.api.formatter import Formatter
from clikit.api.io import IO
from clikit.ui import Component
from clikit.ui.components import BorderUtil
from clikit.ui.components import CellWrapper
from clikit.ui.style import TableStyle
from clikit.utils.string import get_string_length


class Table(Component):
    """
    A table of rows and columns.
    """

    def __init__(self, style=None):  # type: (Optional[TableStyle]) -> None
        self._header_row = []
        self._rows = []
        self._nb_columns = None

        if style is None:
            style = TableStyle.ascii()

        self._style = style

    def set_header_row(self, row):  # type: (List[str]) -> Table
        if self._nb_columns is None:
            self._nb_columns = len(row)
        elif len(row) != self._nb_columns:
            raise ValueError(
                "Expected the header row to contain {} cells, but got {}.".format(
                    self._nb_columns, len(row)
                )
            )

        self._header_row = row

        return self

    def add_row(self, row):  # type: (List[str]) -> Table
        if self._nb_columns is None:
            self._nb_columns = len(row)
        elif len(row) != self._nb_columns:
            raise ValueError(
                "Expected the row to contain {} cells, but got {}.".format(
                    self._nb_columns, len(row)
                )
            )

        self._rows.append(row)

        return self

    def add_rows(self, rows):  # type: (List[List[str]]) -> Table
        for row in rows:
            self.add_row(row)

        return self

    def set_rows(self, rows):  # type: (List[List[str]]) -> Table
        self._rows = []

        for row in rows:
            self.add_row(row)

        return self

    def set_row(self, index, row):  # type: (int, List[str]) -> Table
        if len(row) != self._nb_columns:
            raise ValueError(
                "Expected the row to contain {} cells, but got {}.".format(
                    self._nb_columns, len(row)
                )
            )

        self._rows[index] = row

        return self

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        if not self._rows:
            return

        screen_width = io.terminal_dimensions.width
        excess_column_width = max(
            get_string_length(self._style.header_cell_format.format("")),
            get_string_length(self._style.cell_format.format("")),
        )

        wrapper = self._get_cell_wrapper(
            io, screen_width, excess_column_width, indentation
        )

        return self._render_rows(
            io,
            wrapper.wrapped_rows,
            wrapper.column_lengths,
            excess_column_width,
            indentation,
        )

    def _get_cell_wrapper(
        self, formatter, screen_width, excess_column_width, indentation
    ):  # type: (Formatter, int, int, int) -> CellWrapper
        border_style = self._style.border_style
        border_width = (
            get_string_length(border_style.line_vl_char)
            + (self._nb_columns - 1) * get_string_length(border_style.line_vc_char)
            + get_string_length(border_style.line_vr_char)
        )
        available_width = (
            screen_width
            - indentation
            - border_width
            - self._nb_columns * excess_column_width
        )

        wrapper = CellWrapper()

        for header_cell in self._header_row:
            wrapper.add_cell(header_cell)

        for row in self._rows:
            for cell in row:
                wrapper.add_cell(cell)

        wrapper.fit(available_width, self._nb_columns, formatter)

        return wrapper

    def _render_rows(
        self, io, rows, column_lengths, excess_column_length, indentation
    ):  # type: (IO, List[List[str]], List[int], int, int) -> None
        alignments = self._style.get_column_alignments(len(column_lengths))
        border_style = self._style.border_style
        border_column_lengths = [
            length + excess_column_length for length in column_lengths
        ]

        BorderUtil.draw_top_border(io, border_style, border_column_lengths, indentation)

        if self._header_row:
            BorderUtil.draw_row(
                io,
                border_style,
                rows.pop(0),
                column_lengths,
                alignments,
                self._style.header_cell_format,
                self._style.padding_char,
                self._style.header_cell_style,
                indentation,
            )

            BorderUtil.draw_middle_border(
                io, border_style, border_column_lengths, indentation
            )

        for row in rows:
            BorderUtil.draw_row(
                io,
                border_style,
                row,
                column_lengths,
                alignments,
                self._style.cell_format,
                self._style.padding_char,
                self._style.cell_style,
                indentation,
            )

        BorderUtil.draw_bottom_border(
            io, border_style, border_column_lengths, indentation
        )



from clikit.api.exceptions import CliKitException


class CannotAddOptionException(RuntimeError):
    @classmethod
    def already_exists(cls, name):
        return cls(
            'An option named "{}{}" exists already.'.format(
                "--" if len(name) > 1 else "-", name
            )
        )


class NoSuchOptionException(RuntimeError):
    def __init__(self, name):
        message = 'The "{}{}" option does not exist.'.format(
            "--" if len(name) > 1 else "-", name
        )

        super(NoSuchOptionException, self).__init__(message)


class CannotAddArgumentException(RuntimeError):
    @classmethod
    def already_exists(cls, name):
        return cls('An argument named "{}" exists already.'.format(name))

    @classmethod
    def cannot_add_after_multi_valued(cls):
        return cls("Cannot add an argument after a multi-valued argument.")

    @classmethod
    def cannot_add_required_after_optional(cls):
        return cls("Cannot add a required argument after an optional one.")


class NoSuchArgumentException(RuntimeError):
    def __init__(self, name):
        if isinstance(name, int):
            message = "The argument at position {} does not exist.".format(name)
        else:
            message = 'The "{}" argument does not exist.'.format(name)

        super(NoSuchArgumentException, self).__init__(message)


class CannotParseArgsException(RuntimeError, CliKitException):
    @classmethod
    def too_many_arguments(cls):
        return cls("Too many arguments.")

    @classmethod
    def option_does_not_accept_value(cls, name):
        if len(name) > 1:
            name = "--" + name
        else:
            name = "--" + name

        return cls('The "{}" option does not accept a value.'.format(name))

    @classmethod
    def option_requires_value(cls, name):
        if len(name) > 1:
            name = "--" + name
        else:
            name = "--" + name

        return cls('The "{}" option requires a value.'.format(name))



from clikit.api.io import IO
from clikit.api.io import Input
from clikit.api.io import Output

from .input_stream import NullInputStream
from .output_stream import NullOutputStream


class NullIO(IO):
    """
    An I/O that does nothing.
    """

    def __init__(self):  # type: () -> None
        input = Input(NullInputStream())
        output = Output(NullOutputStream())
        error_output = Output(NullOutputStream())

        super(NullIO, self).__init__(input, output, error_output)



from typing import Dict
from typing import List
from typing import Optional


from clikit.api.args.args import Args
from clikit.api.args.args_parser import ArgsParser
from clikit.api.args.format import ArgsFormat
from clikit.api.args.format import Argument
from clikit.api.args.format import CommandName
from clikit.api.args.exceptions import CannotParseArgsException
from clikit.api.args.exceptions import NoSuchOptionException
from clikit.api.args.raw_args import RawArgs

from clikit.utils._compat import OrderedDict


class DefaultArgsParser(ArgsParser):
    """
    Default parser for RawArgs instances.
    """

    def __init__(self):  # type: () -> None
        self._arguments = OrderedDict()
        self._options = OrderedDict()

    def parse(
        self, args, fmt, lenient=False
    ):  # type: (RawArgs, ArgsFormat, bool) -> Args
        self._arguments = OrderedDict()

        arguments = OrderedDict()
        command_names = OrderedDict()
        i = 1
        for j, command_name in enumerate(fmt.get_command_names()):
            arg_name = "cmd{}{}".format(j + 1, i)
            while fmt.has_argument(arg_name):
                i += 1
                arg_name = "cmd{}{}".format(j + 1, i)

            arguments[arg_name] = Argument(arg_name, Argument.REQUIRED)
            command_names[arg_name] = command_name

        arguments.update(fmt.get_arguments())

        _fmt = ArgsFormat(
            fmt.get_command_names()
            + list(arguments.values())
            + list(fmt.get_options().values())
        )

        try:
            self._parse(args, _fmt, lenient)
        except (CannotParseArgsException, NoSuchOptionException):
            if not lenient:
                raise

        self._insert_missing_command_names(arguments, command_names, lenient)

        # Validate
        missing_arguments = [
            arg.name
            for arg in _fmt.get_arguments().values()
            if arg.name not in self._arguments and arg.is_required()
        ]
        if missing_arguments and not lenient:
            raise CannotParseArgsException(
                'Not enough arguments (missing: "{}").'.format(
                    ", ".join(missing_arguments)
                )
            )

        parsed_args = Args(fmt, args)
        for name, value in self._arguments.items():
            if fmt.has_argument(name):
                parsed_args.set_argument(name, value)

        for name, value in self._options.items():
            if fmt.has_option(name):
                parsed_args.set_option(name, value)

        return parsed_args

    def _parse(
        self, raw_args, fmt, lenient
    ):  # type: (RawArgs, ArgsFormat, bool) -> None
        tokens = raw_args.tokens[:]

        parse_options = True
        while True:
            try:
                token = tokens.pop(0)
            except IndexError:
                break

            if parse_options and token == "":
                self._parse_argument(token, fmt, lenient)
            elif parse_options and token == "--":
                parse_options = False
            elif parse_options and token.find("--") == 0:
                self._parse_long_option(token, tokens, fmt, lenient)
            elif parse_options and token[0] == "-" and token != "-":
                self._parse_short_option(token, tokens, fmt, lenient)
            else:
                self._parse_argument(token, fmt, lenient)

    def _insert_missing_command_names(
        self, arguments, command_names, lenient=False
    ):  # type: (Dict[str, Argument], Dict[str, CommandName], bool) -> None
        fixed_values = {}

        actual_values = self._flatten(self._arguments.values())

        command_names_iterator = iter(command_names.values())
        actual_values_iterator = iter(actual_values)
        arguments_iterator = iter(arguments.values())

        actual_value, command_name = self._skip_command_names(
            actual_values_iterator, command_names_iterator, arguments_iterator
        )

        _, argument = self._copy_argument_values(
            command_names_iterator,
            command_name,
            arguments_iterator,
            None,
            fixed_values,
            lenient,
        )
        self._copy_argument_values(
            actual_values_iterator,
            actual_value,
            arguments_iterator,
            argument,
            fixed_values,
            lenient,
        )

        for name, value in fixed_values.items():
            self._arguments[name] = value

    def _skip_command_names(self, actual_values, command_names, arguments):
        arg = next(actual_values, None)
        command_name = next(command_names, None)

        while arg and command_name and command_name.match(arg):
            arg = next(actual_values, None)
            command_name = next(command_names, None)
            next(arguments, None)

        return arg, command_name

    def _copy_argument_values(
        self, actual_values, value, arguments, argument, fixed_values, lenient
    ):
        if value is None:
            value = next(actual_values, None)

        if argument is None:
            argument = next(arguments, None)

        while value is not None:
            if argument is None:
                if lenient:
                    return

                raise CannotParseArgsException.too_many_arguments()

            name = argument.name

            # Append the value to multi-valued arguments
            if argument.is_multi_valued():
                if name not in fixed_values:
                    fixed_values[name] = []

                fixed_values[name].append(value)

                # The multi-valued argument is the last one, so we don't
                # need to advance the array pointer anymore.
            else:
                fixed_values[name] = value

                argument = next(arguments, None)

            value = next(actual_values, None)

        return value, argument

    def _flatten(self, arguments, result=None):
        if result is None:
            result = []

        for value in arguments:
            if isinstance(value, list):
                self._flatten(value, result)
            else:
                result.append(value)

        return result

    def _parse_argument(
        self, token, fmt, lenient
    ):  # type: (str, ArgsFormat, bool) -> None
        c = len(self._arguments)

        # if input is expecting another argument, add it
        if fmt.has_argument(c):
            arg = fmt.get_argument(c)
            if arg.is_multi_valued():
                if arg.name not in self._arguments:
                    self._arguments[arg.name] = []

                self._arguments[arg.name].append(token)
            else:
                self._arguments[arg.name] = token
        elif fmt.has_argument(c - 1) and fmt.get_argument(c - 1).is_multi_valued():
            arg = fmt.get_argument(c - 1)
            if arg.name not in self._arguments:
                self._arguments[arg.name] = []

            self._arguments[arg.name].append(token)
        # unexpected argument
        else:
            if not lenient:
                raise CannotParseArgsException.too_many_arguments()

    def _parse_long_option(
        self, token, tokens, fmt, lenient
    ):  # type: (str, List[str], ArgsFormat, bool) -> None
        name = token[2:]
        pos = name.find("=")
        if pos != -1:
            self._add_long_option(name[:pos], name[pos + 1 :], tokens, fmt, lenient)
        else:
            if fmt.has_option(name) and fmt.get_option(name).accepts_value():
                try:
                    value = tokens.pop(0)
                except IndexError:
                    value = None

                if value and value.startswith("-"):
                    tokens.insert(0, value)
                    value = None

                self._add_long_option(name, value, tokens, fmt, lenient)
            else:
                self._add_long_option(name, None, tokens, fmt, lenient)

    def _parse_short_option(
        self, token, tokens, fmt, lenient
    ):  # type: (str, List[str], ArgsFormat, bool) -> None
        name = token[1:]
        if len(name) > 1:
            if fmt.has_option(name[0]) and fmt.get_option(name[0]).accepts_value():
                # an option with a value (with no space)
                self._add_short_option(name[0], name[1:], tokens, fmt, lenient)
            else:
                self._parse_short_option_set(name, tokens, fmt, lenient)
        else:
            if fmt.has_option(name[0]) and fmt.get_option(name[0]).accepts_value():
                try:
                    value = tokens.pop(0)
                except IndexError:
                    value = None

                if value and value.startswith("-"):
                    tokens.insert(0, value)
                    value = None

                self._add_short_option(name, value, tokens, fmt, lenient)
            else:
                self._add_short_option(name, None, tokens, fmt, lenient)

    def _parse_short_option_set(
        self, name, tokens, fmt, lenient
    ):  # type: (str, List[str], ArgsFormat, bool) -> None
        l = len(name)
        for i in range(0, l):
            if not fmt.has_option(name[i]):
                raise NoSuchOptionException(name[i])

            option = fmt.get_option(name[i])
            if option.accepts_value():
                self._add_long_option(
                    option.long_name,
                    None if l - 1 == i else name[i + 1 :],
                    tokens,
                    fmt,
                    lenient,
                )

                break
            else:
                self._add_long_option(option.long_name, None, tokens, fmt, lenient)

    def _add_long_option(
        self, name, value, tokens, fmt, lenient
    ):  # type: (str, Optional[str], List[str], ArgsFormat, bool) -> None
        if not fmt.has_option(name):
            raise NoSuchOptionException(name)

        option = fmt.get_option(name)

        if value is False:
            value = None

        if value is not None and not option.accepts_value():
            raise CannotParseArgsException.option_does_not_accept_value(name)

        if value is None and option.accepts_value() and len(tokens):
            # if option accepts an optional or mandatory argument
            # let's see if there is one provided
            try:
                nxt = tokens.pop(0)
            except IndexError:
                nxt = None

            if nxt and len(nxt) >= 1 and nxt[0] != "-":
                value = nxt
            elif not nxt:
                value = ""
            else:
                tokens.insert(0, nxt)

        # This test is here to handle cases like --foo=
        # and foo option value is optional
        if value == "":
            value = None

        if value is None:
            if option.is_value_required():
                raise CannotParseArgsException.option_requires_value(name)

            if not option.is_multi_valued():
                value = option.default if option.is_value_optional() else True

        if option.is_multi_valued():
            if name not in self._options:
                self._options[name] = []

            self._options[name].append(value)
        else:
            self._options[name] = value

    def _add_short_option(
        self, name, value, tokens, fmt, lenient
    ):  # type: (str, Optional[str], List[str], ArgsFormat, bool) -> None
        if not fmt.has_option(name):
            raise NoSuchOptionException(name)

        self._add_long_option(
            fmt.get_option(name).long_name, value, tokens, fmt, lenient
        )



import re

from .question import Question


class ConfirmationQuestion(Question):
    """
    Represents a yes/no question.
    """

    def __init__(self, question, default=True, true_answer_regex="(?i)^y"):
        super(ConfirmationQuestion, self).__init__(question, default)

        self._true_answer_regex = true_answer_regex
        self._normalizer = self._get_default_normalizer

    def _write_prompt(self, io):
        message = self._question

        message = "<question>{} (yes/no)</> [<comment>{}</>] ".format(
            message, "yes" if self._default else "no"
        )

        io.error(message)

    def _get_default_normalizer(self, answer):
        """
        Default answer normalizer.
        """
        if isinstance(answer, bool):
            return answer

        answer_is_true = re.match(self._true_answer_regex, answer) is not None
        if self.default is False:
            return answer and answer_is_true

        return not answer or answer_is_true



from contextlib import contextmanager

from typing import Any
from typing import List
from typing import Optional

from clikit.api.args.args_parser import ArgsParser
from clikit.api.args.format.args_format import ArgsFormat
from clikit.api.args.format.args_format_builder import ArgsFormatBuilder
from clikit.api.args.format.command_name import CommandName
from clikit.api.command.exceptions import NoSuchCommandException

from .config import Config


class CommandConfig(Config):
    """
    The configuration of a console command.
    """

    def __init__(self, name=None):  # type: (Optional[str]) -> None
        super(CommandConfig, self).__init__()

        self._name = name
        self._aliases = []
        self._description = ""
        self._help = None
        self._enabled = True
        self._hidden = False
        self._process_title = None
        self._default = None
        self._anonymous = None
        self._sub_command_configs = []  # type: List[CommandConfig]

        self._parent_config = None  # type: Optional[CommandConfig]

    @property
    def name(self):  # type: () -> Optional[str]
        return self._name

    def set_name(self, name):  # type: (Optional[str]) -> CommandConfig
        self._name = name

        return self

    @property
    def aliases(self):  # type: () -> List[str]
        return self._aliases

    def add_alias(self, alias):  # type: (str) -> CommandConfig
        self._aliases.append(alias)

        return self

    def add_aliases(self, aliases):  # type: (List[str]) -> CommandConfig
        for alias in aliases:
            self.add_alias(alias)

        return self

    def set_aliases(self, aliases):  # type: (List[str]) -> CommandConfig
        self._aliases = []

        return self.add_aliases(aliases)

    @property
    def description(self):  # type: () -> str
        return self._description

    def set_description(self, description):  # type: (str) -> CommandConfig
        self._description = description

        return self

    @property
    def help(self):  # type: () -> Optional[str]
        return self._help

    def set_help(self, help):  # type: (Optional[str]) -> CommandConfig
        self._help = help

        return self

    def is_enabled(self):  # type: () -> bool
        return self._enabled

    def enable(self):  # type: () -> CommandConfig
        self._enabled = True

        return self

    def disable(self):  # type: () -> CommandConfig
        self._enabled = False

        return self

    def is_hidden(self):  # type: () -> bool
        return self._hidden

    def hide(self, hidden=True):  # type: (bool) -> CommandConfig
        self._hidden = hidden

        return self

    @property
    def process_title(self):  # type: () -> Optional[str]
        return self._process_title

    def set_process_title(
        self, process_title
    ):  # type: (Optional[str]) -> CommandConfig
        self._process_title = process_title

        return self

    def default(self, default=True):  # type: (bool) -> CommandConfig
        """
        Marks the command as the default command.
        """
        self._default = default
        self._anonymous = False

        return self

    def anonymous(self):  # type: () -> CommandConfig
        self._default = True
        self._anonymous = True

        return self

    def is_default(self):  # type: () -> bool
        return self._default

    def is_anonymous(self):  # type: () -> bool
        return self._anonymous

    @property
    def parent_config(self):  # type: () -> Optional[CommandConfig]
        return self._parent_config

    def set_parent_config(
        self, parent_config
    ):  # type: (Optional[CommandConfig]) -> CommandConfig
        self._parent_config = parent_config

        return self

    def is_sub_command_config(self):  # type: () -> bool
        return self._parent_config is not None

    def build_args_format(
        self, base_format=None
    ):  # type: (Optional[ArgsFormat]) -> ArgsFormat
        builder = ArgsFormatBuilder(base_format)

        if not self._anonymous:
            builder.add_command_name(CommandName(self.name, self.aliases))

        builder.add_options(*self.options.values())
        builder.add_arguments(*self.arguments.values())

        return builder.format

    @contextmanager
    def sub_command(self, name):  # type: (str) -> CommandConfig
        sub_command_config = CommandConfig(name)
        self.add_sub_command_config(sub_command_config)

        yield sub_command_config

    def create_sub_command(self, name):  # type: (str) -> CommandConfig
        sub_command_config = CommandConfig(name)
        self.add_sub_command_config(sub_command_config)

        return sub_command_config

    @contextmanager
    def edit_sub_command(self, name):  # type: (str) -> CommandConfig
        sub_command_config = self.get_sub_command_config(name)

        yield sub_command_config

    def add_sub_command_config(
        self, sub_command_config
    ):  # type: (CommandConfig) -> CommandConfig
        self._sub_command_configs.append(sub_command_config)

        return self

    def add_sub_command_configs(
        self, sub_command_configs
    ):  # type: (List[CommandConfig]) -> CommandConfig
        for sub_command_config in sub_command_configs:
            self.add_sub_command_config(sub_command_config)

        return self

    def get_sub_command_config(self, name):  # type: (str) -> CommandConfig
        for sub_command_config in self._sub_command_configs:
            if sub_command_config.name == name:
                return sub_command_config

        raise NoSuchCommandException(name)

    @property
    def sub_command_configs(self):  # type: () -> List[CommandConfig]
        return self._sub_command_configs

    def has_sub_command_config(self, name):  # type: (str) -> bool
        for sub_command_config in self._sub_command_configs:
            if sub_command_config.name == name:
                return True

        return False

    def has_sub_command_configs(self):  # type: () -> bool
        return len(self._sub_command_configs) > 0

    @property
    def default_helper_set(self):  # type: () -> HelperSet
        if self._parent_config:
            return self._parent_config.default_helper_set

        return super(CommandConfig, self).default_helper_set

    @property
    def default_args_parser(self):  # type: () -> ArgsParser
        if self._parent_config:
            return self._parent_config.default_args_parser

        return super(CommandConfig, self).default_args_parser

    @property
    def default_lenient_args_parsing(self):  # type: () -> bool
        if self._parent_config:
            return self._parent_config.default_lenient_args_parsing

        return super(CommandConfig, self).default_lenient_args_parsing

    @property
    def default_handler(self):  # type: () -> Any
        if self._parent_config:
            return self._parent_config.default_handler

        return super(CommandConfig, self).default_handler

    @property
    def default_handler_method(self):  # type: () -> str
        if self._parent_config:
            return self._parent_config.default_handler_method

        return super(CommandConfig, self).default_handler_method



import io

from typing import Optional

from clikit.api.io.input_stream import InputStream


class StreamInputStream(InputStream):
    """
    An input stream that reads from a stream.
    """

    def __init__(self, stream):  # type: (io.TextIOWrapper) -> None
        self._stream = stream

        if hasattr("stream", "seekable") and stream.seekable():
            stream.seek(0)

    def read(self, length):  # type: (int) -> str
        """
        Reads the given amount of characters from the stream.
        """
        if self.is_closed():
            raise io.UnsupportedOperation("Cannot read from a closed input.")

        try:
            data = self._stream.read(length)
        except EOFError:
            return ""

        if data:
            return data

        return ""

    def read_line(self, length=None):  # type: (Optional[int]) -> str
        """
        Reads a line from the stream.
        """
        if self.is_closed():
            raise io.UnsupportedOperation("Cannot read from a closed input.")

        try:
            return self._stream.readline(length) or ""
        except EOFError:
            return ""

    def close(self):  # type: () -> None
        """
        Closes the stream
        """
        self._stream.close()

    def is_closed(self):  # type: () -> bool
        """
        Returns whether the stream is closed or not
        """
        return self._stream.closed



import re

from contextlib import contextmanager

from typing import Callable
from typing import ContextManager
from typing import List
from typing import Optional

from clikit.api.event import EventDispatcher
from clikit.api.formatter import Style
from clikit.api.formatter import StyleSet
from clikit.api.command.exceptions import NoSuchCommandException

from .command_config import CommandConfig
from .config import Config


class ApplicationConfig(Config):
    """
    The configuration of a console application.
    """

    def __init__(
        self, name=None, version=None
    ):  # type: (Optional[str], Optional[str]) -> None
        self._name = name
        self._version = version
        self._display_name = None
        self._help = None
        self._command_configs = []  # type: List[CommandConfig]
        self._catch_exceptions = True
        self._terminate_after_run = True
        self._command_resolver = None
        self._io_factory = None
        self._debug = False
        self._style_set = None
        self._dispatcher = None
        self._pre_resolve_hooks = []  # type: List[Callable]

        super(ApplicationConfig, self).__init__()

    @property
    def name(self):  # type: () -> Optional[str]
        return self._name

    def set_name(self, name):  # type: (Optional[str]) -> ApplicationConfig
        self._name = name

        return self

    @property
    def display_name(self):  # type: () -> str
        """
        Returns the application name as it is displayed in the help.
        """
        if self._display_name is not None:
            return self._display_name

        return self.default_display_name

    def set_display_name(
        self, display_name
    ):  # type: (Optional[str]) -> ApplicationConfig
        self._display_name = display_name

        return self

    @property
    def version(self):  # type: () -> Optional[str]
        return self._version

    def set_version(self, version):  # type: (Optional[str]) -> ApplicationConfig
        self._version = version

        return self

    @property
    def help(self):  # type: () -> Optional[str]
        return self._help

    def set_help(self, help):  # type: (Optional[str]) -> ApplicationConfig
        self._help = help

        return self

    @property
    def dispatcher(self):  # type: () -> EventDispatcher
        return self._dispatcher

    def set_event_dispatcher(
        self, dispatcher
    ):  # type: (EventDispatcher) -> ApplicationConfig
        self._dispatcher = dispatcher

        return self

    def add_event_listener(
        self, event_name, listener, priority=0
    ):  # type: (str, Callable, int) -> ApplicationConfig
        if self._dispatcher is None:
            self._dispatcher = EventDispatcher()

        self._dispatcher.add_listener(event_name, listener, priority)

        return self

    def is_exception_caught(self):  # type: () -> bool
        return self._catch_exceptions

    def set_catch_exceptions(
        self, catch_exceptions
    ):  # type: (bool) -> ApplicationConfig
        self._catch_exceptions = catch_exceptions

        return self

    def is_terminated_after_run(self):  # type: () -> bool
        return self._terminate_after_run

    def set_terminate_after_run(
        self, terminate_after_run
    ):  # type: (bool) -> ApplicationConfig
        self._terminate_after_run = terminate_after_run

        return self

    @property
    def command_resolver(self):  # type: () -> CommandResolver
        if self._command_resolver is None:
            self._command_resolver = self.default_command_resolver

        return self._command_resolver

    def set_command_resolver(
        self, command_resolver
    ):  # type: (CommandResolver) -> ApplicationConfig
        self._command_resolver = command_resolver

        return self

    @property
    def io_factory(self):  # type: () -> Optional[Callable]
        return self._io_factory

    def set_io_factory(
        self, io_factory
    ):  # type: (Optional[Callable]) -> ApplicationConfig
        self._io_factory = io_factory

        return self

    def is_debug(self):  # type: () -> bool
        return self._debug

    def debug(self, debug=True):  # type: (bool) -> ApplicationConfig
        self._debug = debug

        return self

    @property
    def style_set(self):  # type: () -> StyleSet
        if self._style_set is None:
            self._style_set = self.default_style_set

        return self._style_set

    def set_style_set(self, style_set):  # type: (StyleSet) -> ApplicationConfig
        self._style_set = style_set

        return self

    def add_style(self, style):  # type: (Style) -> ApplicationConfig
        self.style_set.add(style)

        return self

    def add_styles(self, styles):  # type: (List[Style]) -> ApplicationConfig
        for style in styles:
            self.add_style(style)

        return self

    def remove_style(self, tag):  # type: (str) -> ApplicationConfig
        self.style_set.remove(tag)

    @contextmanager
    def command(self, name):  # type: (str) -> ContextManager[CommandConfig]
        command_config = CommandConfig(name)
        self.add_command_config(command_config)

        yield command_config

    def create_command(self, name):  # type: (str) -> CommandConfig
        command_config = CommandConfig(name)
        self.add_command_config(command_config)

        return command_config

    @contextmanager
    def edit_command(self, name):  # type: (str) -> CommandConfig
        command_config = self.get_command_config(name)

        yield command_config

    def add_command_config(
        self, command_config
    ):  # type: (CommandConfig) -> ApplicationConfig
        self._command_configs.append(command_config)

        return self

    def add_command_configs(
        self, command_configs
    ):  # type: (List[CommandConfig]) -> ApplicationConfig
        for command_config in command_configs:
            self.add_command_config(command_config)

        return self

    def get_command_config(self, name):  # type: (str) -> CommandConfig
        for command_config in self._command_configs:
            if command_config.name == name:
                return command_config

        raise NoSuchCommandException(name)

    @property
    def command_configs(self):  # type: () -> List[CommandConfig]
        return self._command_configs

    def has_command_config(self, name):  # type: (str) -> bool
        for command_config in self._command_configs:
            if command_config.name == name:
                return True

        raise False

    def has_command_configs(self):  # type: () -> bool
        return len(self._command_configs) > 0

    @property
    def default_display_name(self):  # type: () -> Optional[str]
        if self._name is None:
            return

        return re.sub(r"[\s\-_]+", " ", self._name).title()

    @property
    def default_style_set(self):  # type: () -> StyleSet
        raise NotImplementedError()

    @property
    def default_command_resolver(self):
        raise NotImplementedError()



from typing import Any
from typing import Dict
from typing import Optional

from clikit.api.args.args_parser import ArgsParser
from clikit.api.args.format.args_format_builder import ArgsFormatBuilder
from clikit.api.args.format.argument import Argument
from clikit.api.args.format.option import Option
from clikit.args.default_args_parser import DefaultArgsParser


class Config(object):
    """
    Implements methods shared by all configurations.
    """

    def __init__(self):
        self._format_builder = ArgsFormatBuilder()
        self._helper_set = None
        self._args_parser = None
        self._lenient_args_parsing = None
        self._handler = None
        self._handler_method = None

        self.configure()

    @property
    def arguments(self):  # type: () -> Dict[str, Argument]
        return self._format_builder.get_arguments()

    def add_argument(
        self, name, flags=0, description=None, default=None
    ):  # type: (str, int, Optional[str], Any) -> Config
        argument = Argument(name, flags, description, default)

        self._format_builder.add_argument(argument)

        return self

    @property
    def options(self):  # type: () -> Dict[str, Option]
        return self._format_builder.get_options()

    def add_option(
        self,
        long_name,
        short_name=None,
        flags=0,
        description=None,
        default=None,
        value_name="...",
    ):  # type: (str, Optional[str], int, Optional[str], Any, str) -> Config
        option = Option(long_name, short_name, flags, description, default, value_name)

        self._format_builder.add_option(option)

        return self

    @property
    def helper_set(self):  # type: () -> HelperSet
        if self._helper_set is None:
            return self.default_helper_set

        return self._helper_set

    def set_helper_set(self, helper_set):  # type: (HelperSet) -> Config
        self._helper_set = helper_set

        return self

    @property
    def args_parser(self):  # type: () -> ArgsParser
        if self._args_parser is None:
            return self.default_args_parser

        return self._args_parser

    def set_args_parser(self, args_parser):  # type: (ArgsParser) -> Config
        self._args_parser = args_parser

        return self

    def is_lenient_args_parsing_enabled(self):  # type: () -> bool
        if self._lenient_args_parsing is None:
            return self.default_lenient_args_parsing

        return self._lenient_args_parsing

    def enable_lenient_args_parsing(self):  # type: () -> Config
        self._lenient_args_parsing = True

        return self

    def disable_lenient_args_parsing(self):  # type: () -> Config
        self._lenient_args_parsing = False

        return self

    @property
    def handler(self):  # type: () -> Any
        if self._handler is None:
            return self.default_handler

        if callable(self._handler):
            return self._handler()

        return self._handler

    def set_handler(self, handler):  # type: (Any) -> Config
        self._handler = handler

        return self

    @property
    def handler_method(self):  # type: () -> str
        if self._handler_method is None:
            return self.default_handler_method

        return self._handler_method

    def set_handler_method(self, handler_method):  # type: (str) -> Config
        self._handler_method = handler_method

        return self

    @property
    def default_helper_set(self):  # type: () -> HelperSet
        return HelperSet()

    @property
    def default_args_parser(self):  # type: () -> ArgsParser
        return DefaultArgsParser()

    @property
    def default_lenient_args_parsing(self):  # type: () -> bool
        return False

    @property
    def default_handler(self):  # type: () -> Any
        return lambda: None

    @property
    def default_handler_method(self):  # type: () -> Any
        return "handle"

    def configure(self):  # type: () -> None
        """
        Adds the default configuration.

        Should be overridden in subclasses
        """



from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from .format.args_format import ArgsFormat
from .raw_args import RawArgs


class Args(object):
    """
    The parsed console arguments.
    """

    def __init__(self, fmt, raw_args=None):  # type: (ArgsFormat, RawArgs) -> None
        self._fmt = fmt
        self._raw_args = raw_args
        self._options = {}
        self._arguments = {}

    @property
    def format(self):  # type: () -> ArgsFormat
        return self._fmt

    @property
    def raw_args(self):  # type: () -> RawArgs
        return self._raw_args

    @property
    def script_name(self):  # type: () -> Optional[str]
        if self._raw_args:
            return self._raw_args.script_name

    @property
    def command_names(self):
        return self._fmt.get_command_names()

    @property
    def command_options(self):
        return self._fmt.get_command_options()

    def option(self, name):
        option = self._fmt.get_option(name)

        if option.long_name in self._options:
            return self._options[option.long_name]

        if option.accepts_value():
            return option.default

        return False

    def options(self, include_defaults=True):
        options = self._options.copy()

        if include_defaults:
            for option in self._fmt.get_options().values():
                name = option.long_name

                if not name in options:
                    default = False
                    if option.accepts_value():
                        default = option.default

                    options[name] = default

        return options

    def set_option(self, name, value=True):
        option = self._fmt.get_option(name)

        if option.is_multi_valued():
            if not isinstance(value, list):
                value = [value]

            for i, v in enumerate(value):
                value[i] = option.parse(v)
        elif option.accepts_value():
            value = option.parse(value)
        elif value is False:
            if option.long_name in self._options:
                del self._options[option.long_name]

            return self
        else:
            value = True

        self._options[option.long_name] = value

        return self

    def is_option_set(self, name):  # type: (str) -> bool
        return name in self._options

    def is_option_defined(self, name):  # type: (str) -> bool
        return self._fmt.has_option(name)

    def argument(self, name):  # type: (Union[str, int]) -> Any
        argument = self._fmt.get_argument(name)

        if argument.name in self._arguments:
            return self._arguments[name]

        return argument.default

    def arguments(self, include_defaults=True):  # type: (bool) -> Dict[str, Any]
        arguments = {}

        for argument in self._fmt.get_arguments().values():
            name = argument.name

            if name in self._arguments:
                arguments[name] = self._arguments[name]
            elif include_defaults:
                arguments[name] = argument.default

        return arguments

    def set_argument(self, name, value):  # type: (Union[str, int], Any) -> Args
        argument = self._fmt.get_argument(name)

        if argument.is_multi_valued():
            if not isinstance(value, list):
                value = [value]

            for i, v in enumerate(value):
                value[i] = argument.parse(v)
        else:
            value = argument.parse(value)

        self._arguments[argument.name] = value

        return self

    def is_argument_set(self, name):  # type: (Union[str, int]) -> bool
        return name in self._arguments

    def is_argument_defined(self, name):  # type: (Union[str, int]) -> bool
        return self._fmt.has_argument(name)



import getpass
import os
import subprocess

from typing import Any

from clikit.api.formatter import Style
from clikit.api.io import IO
from clikit.utils._compat import decode


class Question(object):
    """
    A question that will be asked in a Console.
    """

    def __init__(self, question, default=None):  # type: (str, Any) -> None
        self._question = question
        self._default = default

        self._attempts = None
        self._hidden = False
        self._hidden_fallback = True
        self._autocomplete_values = None
        self._validator = None
        self._normalizer = None
        self._error_message = None

    @property
    def question(self):  # type: () -> str
        return self._question

    @property
    def default(self):  # type: () -> Any
        return self._default

    @property
    def autocomplete_values(self):
        return self._autocomplete_values

    @property
    def max_attempts(self):
        return self._attempts

    def is_hidden(self):  # type: () -> bool
        return self._hidden

    def hide(self, hidden=True):  # type: (bool) -> None
        if hidden is True and self._autocomplete_values:
            raise RuntimeError("A hidden question cannot use the autocompleter.")

        self._hidden = hidden

    def set_autocomplete_values(self, autocomplete_values):
        if self.is_hidden():
            raise RuntimeError("A hidden question cannot use the autocompleter.")

        self._autocomplete_values = autocomplete_values

    def set_max_attempts(self, attempts):
        self._attempts = attempts

    def set_validator(self, validator):
        self._validator = validator

    def ask(self, io):  # type: (IO) -> str
        """
        Asks the question to the user.
        """
        if not io.is_interactive():
            return self.default

        if not self._validator:
            return self._do_ask(io)

        interviewer = lambda: self._do_ask(io)

        return self._validate_attempts(interviewer, io)

    def _do_ask(self, io):  # type: (IO) -> str
        """
        Asks the question to the user.
        """
        self._write_prompt(io)

        autocomplete = self._autocomplete_values

        if autocomplete is None or not self._has_stty_available():
            ret = False

            if self.is_hidden():
                try:
                    ret = self._get_hidden_response(io)
                except RuntimeError:
                    if not self._hidden_fallback:
                        raise

            if not ret:
                ret = self._read_from_input(io)
        else:
            ret = self._autocomplete(io)

        if len(ret) <= 0:
            ret = self._default

        if self._normalizer:
            return self._normalizer(ret)

        return ret

    def _write_prompt(self, io):
        """
        Outputs the question prompt.
        """
        message = self._question

        io.error("<question>{}</> ".format(message))

    def _write_error(self, io, error):
        """
        Outputs an error message.
        """
        message = "<error>{}</error>".format(decode(str(error)))

        io.error_line(message)

    def _autocomplete(self, io):  # type: (IO) -> str
        """
        Autocomplete a question.
        """
        autocomplete = self._autocomplete_values

        ret = ""

        i = 0
        ofs = -1
        matches = [x for x in autocomplete]
        num_matches = len(matches)

        stty_mode = decode(subprocess.check_output(["stty", "-g"])).rstrip("\n")

        # Disable icanon (so we can read each keypress) and echo (we'll do echoing here instead)
        subprocess.check_output(["stty", "-icanon", "-echo"])

        # Add highlighted text style
        style = Style("hl").fg("black").bg("white")
        io.error_output.formatter.add_style(style)

        # Read a keypress
        while True:
            c = io.read(1)

            # Backspace character
            if c == "\177":
                if num_matches == 0 and i != 0:
                    i -= 1
                    # Move cursor backwards
                    io.error("\033[1D")

                if i == 0:
                    ofs = -1
                    matches = [x for x in autocomplete]
                    num_matches = len(matches)
                else:
                    num_matches = 0

                # Pop the last character off the end of our string
                ret = ret[:i]
            # Did we read an escape sequence
            elif c == "\033":
                c += io.read(2)

                # A = Up Arrow. B = Down Arrow
                if c[2] == "A" or c[2] == "B":
                    if c[2] == "A" and ofs == -1:
                        ofs = 0

                    if num_matches == 0:
                        continue

                    ofs += -1 if c[2] == "A" else 1
                    ofs = (num_matches + ofs) % num_matches
            elif ord(c) < 32:
                if c == "\t" or c == "\n":
                    if num_matches > 0 and ofs != -1:
                        ret = matches[ofs]
                        # Echo out remaining chars for current match
                        io.error(ret[i:])
                        i = len(ret)

                    if c == "\n":
                        io.error(c)
                        break

                    num_matches = 0

                continue
            else:
                io.error(c)
                ret += c
                i += 1

                num_matches = 0
                ofs = 0

                for value in autocomplete:
                    # If typed characters match the beginning chunk of value (e.g. [AcmeDe]moBundle)
                    if value.startswith(ret) and i != len(value):
                        num_matches += 1
                        matches[num_matches - 1] = value

            # Erase characters from cursor to end of line
            io.error("\033[K")

            if num_matches > 0 and ofs != -1:
                # Save cursor position
                io.error("\0337")
                # Write highlighted text
                io.error("<hl>" + matches[ofs][i:] + "</hl>")
                # Restore cursor position
                io.error("\0338")

        subprocess.call(["stty", "{}".format(decode(stty_mode))])

        return ret

    def _get_hidden_response(self, io):  # type: (IO) -> str
        """
        Gets a hidden response from user.
        """
        return getpass.getpass("", stream=io.error_output.stream)

    def _validate_attempts(self, interviewer, io):  # type: (Callable, IO) -> str
        """
        Validates an attempt.
        """
        error = None
        attempts = self._attempts

        while attempts is None or attempts:
            if error is not None:
                self._write_error(io, error)

            try:
                return self._validator(interviewer())
            except Exception as e:
                error = e

            if attempts is not None:
                attempts -= 1

        raise error

    def _read_from_input(self, io):
        """
        Read user input.
        """
        ret = io.read_line(4096)

        if not ret:
            raise RuntimeError("Aborted")

        return decode(ret.strip())

    def _has_stty_available(self):
        devnull = open(os.devnull, "w")

        try:
            exit_code = subprocess.call(["stty"], stdout=devnull, stderr=devnull)
        except Exception:
            exit_code = 2

        return exit_code == 0



from typing import List

from pylev import levenshtein

from clikit.api.command import CommandCollection


def find_similar_command_names(
    name, commands
):  # type: (str, CommandCollection) -> List[str]
    """
    Finds names similar to a given command name.
    """
    threshold = 1e3
    distance_by_name = {}
    suggested_names = []

    # Include aliases in the search
    actual_names = commands.get_names(True)

    for actual_name in actual_names:
        # Get Levenshtein distance between the input and each command name
        distance = levenshtein(name, actual_name)

        is_similar = distance <= len(name) / 3
        is_sub_string = actual_name.find(name) != -1

        if is_similar or is_sub_string:
            distance_by_name[actual_name] = (
                distance,
                actual_name.find(name) if is_sub_string else float("inf"),
            )

    # Only keep results with a distance below the threshold
    distance_by_name = {
        k: v for k, v in distance_by_name.items() if v[0] < 2 * threshold
    }

    # Display results with shortest distance first
    for k, v in sorted(distance_by_name.items(), key=lambda i: (i[1][0], i[1][1])):
        if k not in suggested_names:
            suggested_names.append(k)

    return suggested_names



import math

from typing import List
from typing import Optional

from clikit.api.formatter import Formatter
from clikit.utils.terminal import Terminal

from .output import Output
from .output_stream import OutputStream


class SectionOutput(Output):
    def __init__(
        self, stream, sections, formatter=None
    ):  # type: (OutputStream, List[SectionOutput], Optional[Formatter]) -> None
        super(SectionOutput, self).__init__(stream, formatter=formatter)

        self._content = []
        self._lines = 0
        sections.insert(0, self)
        self._sections = sections
        self._terminal = Terminal()

    @property
    def content(self):  # type: () -> str
        return "".join(self._content)

    @property
    def lines(self):  # type: () -> int
        return self._lines

    def clear(self, lines=None):  # type: (Optional[int]) -> None
        if (
            not self._content
            or not self.supports_ansi()
            and not self._formatter.force_ansi()
        ):
            return

        if lines:
            # Multiply lines by 2 to cater for each new line added between content
            del self._content[-(lines * 2) :]
        else:
            lines = self._lines
            self._content = []

        self._lines -= lines

        super(SectionOutput, self).write(
            self._pop_stream_content_until_current_section(lines)
        )

    def overwrite(self, message):  # type: (str) -> None
        self.clear()
        self.write_line(message)

    def add_content(self, content):  # type: (str) -> None
        for line_content in content.split("\n"):
            self._lines += (
                math.ceil(
                    len(self.remove_format(line_content).replace("\t", "        "))
                    / self._terminal.width
                )
                or 1
            )
            self._content.append(line_content)
            self._content.append("\n")

    def write(
        self, string, flags=None, new_line=False
    ):  # type: (str, Optional[int], bool) -> None
        if not self.supports_ansi() and not self._formatter.force_ansi():
            return super(SectionOutput, self).write(string, flags=flags)

        erased_content = self._pop_stream_content_until_current_section()

        self.add_content(string)

        super(SectionOutput, self).write(string, new_line=True)
        super(SectionOutput, self).write(erased_content)

    def _pop_stream_content_until_current_section(
        self, lines_to_clear_count=0
    ):  # type: (int) -> str
        erased_content = []

        for section in self._sections:
            if section is self:
                break

            lines_to_clear_count += section.lines
            erased_content.append(section.content)

        if lines_to_clear_count > 0:
            # Move cursor up n lines
            super(SectionOutput, self).write("\x1b[{}A".format(lines_to_clear_count))
            # Erase to end of screen
            super(SectionOutput, self).write("\x1b[0J")

        return "".join(reversed(erased_content))



import itertools
import sys

from typing import List
from typing import Optional


from clikit.api.args.raw_args import RawArgs


class ArgvArgs(RawArgs):
    """
    Console arguments passed via sys.argv.
    """

    def __init__(self, argv=None):  # type: (Optional[List[str]]) -> None
        if argv is None:
            argv = list(sys.argv)

        argv = argv[:]
        self._script_name = argv.pop(0)
        self._tokens = argv
        self._option_tokens = list(
            itertools.takewhile(lambda arg: arg != "--", self.tokens)
        )

    @property
    def script_name(self):  # type: () -> str
        return self._script_name

    @property
    def tokens(self):  # type: () -> List[str]
        return self._tokens

    @property
    def option_tokens(self):  # type: () -> List[str]
        return self._option_tokens

    def has_token(self, token):  # type: (str) -> bool
        return token in self._tokens

    def has_option_token(self, token):  # type: (str) -> bool
        return token in self._option_tokens

    def to_string(self, script_name=True):  # type: (bool) -> str
        string = " ".join(self._tokens)

        if script_name:
            string = self._script_name.lstrip() + " " + string

        return string



from __future__ import division

import textwrap

from typing import List

from clikit.api.formatter import Formatter
from clikit.utils.string import get_max_line_length
from clikit.utils.string import get_max_word_length
from clikit.utils.string import get_string_length


class CellWrapper:
    """
    Wraps cells to fit a given screen width with a given number of columns.
    """

    def __init__(self):  # type: () -> None
        self._cells = []
        self._cell_lengths = []
        self._wrapped_rows = []
        self._nb_columns = 0
        self._column_lengths = []
        self._word_wraps = False
        self._word_cuts = False
        self._max_total_width = 0
        self._total_width = 0

    @property
    def cells(self):  # type: () -> List[str]
        return self._cells

    @property
    def wrapped_rows(self):  # type: () -> List[List[str]]
        return self._wrapped_rows

    @property
    def column_lengths(self):  # type: () -> List[int]
        return self._column_lengths

    @property
    def nb_columns(self):  # type: () -> int
        return self._nb_columns

    @property
    def max_total_width(self):  # type: () -> int
        return self._max_total_width

    @property
    def total_width(self):  # type: () -> int
        return self._total_width

    def add_cell(self, cell):  # type: (str) -> CellWrapper
        self._cells.append(cell.rstrip())

        return self

    def add_cells(self, cells):  # type: (List[str]) -> CellWrapper
        for cell in cells:
            self.add_cell(cell)

        return self

    def get_estimated_nb_columns(self, max_total_width):  # type: (int) -> int
        """
        Returns an estimated number of columns for the given maximum width.
        """
        row_width = 0

        for i, cell in enumerate(self._cells):
            row_width += get_string_length(cell)

            if row_width > max_total_width:
                return i

        return len(self._cells) - 1

    def has_word_wraps(self):  # type: () -> bool
        return self._word_wraps

    def has_word_cuts(self):  # type: () -> bool
        return self._word_cuts

    def fit(
        self, max_total_width, nb_columns, formatter
    ):  # type: (int, int, Formatter) -> None
        self._reset_state(max_total_width, nb_columns)
        self._init_rows(formatter)

        # If the cells fit within the max width we're good
        if self._total_width <= max_total_width:
            return

        self._wrap_columns(formatter)

    def _reset_state(self, max_total_width, nb_columns):  # type: (int, int) -> None
        self._wrapped_rows = []
        self._nb_columns = nb_columns
        self._cell_lengths = []
        self._column_lengths = [0] * nb_columns
        self._word_wraps = False
        self._word_cuts = False
        self._max_total_width = max_total_width
        self._total_width = 0

    def _init_rows(self, formatter):  # type: (Formatter) -> None
        col = 0

        for i, cell in enumerate(self._cells):
            if col == 0:
                self._wrapped_rows.append([""] * self._nb_columns)
                self._cell_lengths.append([0] * self._nb_columns)

            self._wrapped_rows[-1][col] = cell
            self._cell_lengths[-1][col] = get_string_length(cell, formatter)
            self._column_lengths[col] = max(
                self._column_lengths[col], self._cell_lengths[-1][col]
            )

            col = (col + 1) % self._nb_columns

        # Fill last row
        if col > 0:
            while col < self._nb_columns:
                self._wrapped_rows[-1][col] = ""
                self._cell_lengths[-1][col] = 0
                col += 1

        self._total_width = sum(self._column_lengths)

    def _wrap_columns(self, formatter):  # type: (Formatter) -> None
        available_width = self._max_total_width
        long_column_lengths = self._column_lengths[:]

        # Filter "short" column, i.e. columns that are not wrapped
        # We distribute the available screen width by the number of columns
        # and decide that all columns that are shorter than their share are
        # "short".
        # This process is repeated until no more "short" columns are found.
        repeat = True
        while repeat:
            threshold = available_width / len(long_column_lengths)
            repeat = False

            for col, length in enumerate(long_column_lengths):
                if length is not None and length <= threshold:
                    available_width -= length
                    long_column_lengths[col] = None
                    repeat = True

        # Calculate actual and available width
        actual_width = 0
        last_adapted_col = 0

        # "Long" columns, i.e. columns that need to be wrapped, are added to
        # the actual width
        for col, length in enumerate(long_column_lengths):
            if length is None:
                continue

            actual_width += length
            last_adapted_col = col

        # Fit columns into available width
        for col, length in enumerate(long_column_lengths):
            if length is None:
                continue

            # Keep ratios of column lengths and distribute them among the
            # available width
            self._column_lengths[col] = int(
                round((length / actual_width) * available_width)
            )

            if col == last_adapted_col:
                # Fix rounding errors
                self._column_lengths[col] += self._max_total_width - sum(
                    self._column_lengths
                )

            self._wrap_column(col, self._column_lengths[col], formatter)

            # Recalculate the column length based on the actual wrapped length
            self._refresh_column_length(col)

            # Recalculate the actual width based on the changed length.
            actual_width = actual_width - length + self._column_lengths[col]

        self._total_width = sum(self._column_lengths)

    def _wrap_column(
        self, col, column_length, formatter
    ):  # type: (int, List[int], Formatter) -> None
        for i, row in enumerate(self._wrapped_rows):
            cell = row[col]
            cell_length = self._cell_lengths[i][col]

            if cell_length > column_length:
                self._word_wraps = True

                if not self._word_cuts:
                    min_length_without_cut = get_max_word_length(cell, formatter)

                    if min_length_without_cut > column_length:
                        self._word_cuts = True

                # TODO: use format aware wrapper
                wrapped_cell = "\n".join(textwrap.wrap(cell, column_length))

                self._wrapped_rows[i][col] = wrapped_cell

                # Refresh cell length
                self._cell_lengths[i][col] = get_max_line_length(
                    wrapped_cell, formatter
                )

    def _refresh_column_length(self, col):  # type: (int) -> None
        self._column_lengths[col] = 0

        for i, row in enumerate(self._wrapped_rows):
            self._column_lengths[col] = max(
                self._column_lengths[col], self._cell_lengths[i][col]
            )



import ast
import inspect
import keyword
import sys
import traceback

from clikit.api.io import IO
from clikit.utils._compat import PY2


class ExceptionTrace(object):
    """
    Renders the trace of an exception.
    """

    THEME = {
        "comment": "<fg=black;options=bold>",
        "keyword": "<fg=yellow>",
        "builtin": "<fg=blue>",
        "literal": "<fg=magenta>",
    }

    AST_ELEMENTS = {
        "builtins": __builtins__.keys()
        if type(__builtins__) is dict
        else dir(__builtins__),
        "keywords": [
            getattr(ast, cls)
            for cls in dir(ast)
            if keyword.iskeyword(cls.lower())
            and inspect.isclass(getattr(ast, cls))
            and issubclass(getattr(ast, cls), ast.AST)
        ],
    }

    def __init__(self, exception):  # type: (Exception) -> None
        self._exception = exception
        self._exc_info = sys.exc_info()

    def render(self, io, simple=False):  # type: (IO, bool) -> None
        if hasattr(self._exception, "__traceback__"):
            tb = self._exception.__traceback__
        else:
            tb = self._exc_info[2]

        title = ""
        if not simple:
            title += "\n[<error>{}</error>]\n".format(
                self._exception.__class__.__name__
            )

        title += "<error>{}</error>".format(str(self._exception))

        io.write_line(title)

        if not simple and io.is_verbose():
            io.write_line("")
            self._render_traceback(io, tb)

    def _render_traceback(self, io, tb):  # type: (IO, ...) -> None
        frames = []
        while tb:
            frames.append(self._format_traceback_frame(io, tb))

            tb = tb.tb_next

        io.write_line("<b>Traceback (most recent call last):</b>")
        io.write_line("".join(traceback.format_list(frames)))

    def _format_traceback_frame(self, io, tb):  # type: (IO, ...) -> Tuple[Any]
        frame_info = inspect.getframeinfo(tb)
        filename = frame_info.filename
        lineno = frame_info.lineno
        function = frame_info.function
        line = frame_info.code_context[0]

        stripped_line = line.lstrip(" ")
        try:
            tree = ast.parse(stripped_line, mode="exec")
            formatted = self._format_tree(tree, stripped_line, io)
            formatted = (len(line) - len(stripped_line)) * " " + formatted
        except SyntaxError:
            formatted = line

        return (
            io.format("<c1>{}</c1>".format(filename)),
            "<fg=blue;options=bold>{}</>".format(lineno) if not PY2 else lineno,
            "<b>{}</b>".format(function),
            formatted,
        )

    def _format_tree(self, tree, source, io):
        offset = 0
        chunks = []

        nodes = [n for n in ast.walk(tree)]

        displayed_nodes = []

        for node in nodes:
            nodecls = node.__class__
            nodename = nodecls.__name__

            if "col_offset" not in dir(node):
                continue

            if nodecls in self.AST_ELEMENTS["keywords"]:
                displayed_nodes.append((node, nodename.lower(), "keyword"))
            elif nodecls == ast.Name and node.id in self.AST_ELEMENTS["builtins"]:
                displayed_nodes.append((node, node.id, "builtin"))
            elif nodecls == ast.Str:
                displayed_nodes.append((node, "'{}'".format(node.s), "literal"))
            elif nodecls == ast.Num:
                displayed_nodes.append((node, str(node.n), "literal"))

        displayed_nodes.sort(key=lambda elem: elem[0].col_offset)

        for dn in displayed_nodes:
            node = dn[0]
            s = dn[1]
            theme = dn[2]

            begin_col = node.col_offset

            src_chunk = source[offset:begin_col]
            chunks.append(src_chunk)
            chunks.append(io.format("{}{}</>".format(self.THEME[theme], s)))
            offset = begin_col + len(s)

        chunks.append(source[offset:])

        return "".join(chunks)



from clikit.api.args import Args
from clikit.api.io import IO

from .event import Event


class PreHandleEvent(Event):
    """
    Dispatched before a command is handled.

    Add a listener for this event to execute custom logic before or instead of
    the default handler.
    """

    def __init__(self, args, io, command):  # type: (Args, IO, Command) -> None
        super(PreHandleEvent, self).__init__()

        self._args = args
        self._io = io
        self._command = command
        self._handled = False
        self._status_code = 0

    @property
    def args(self):  # type: () -> Args
        return self._args

    @property
    def io(self):  # type: () -> IO
        return self._io

    @property
    def command(self):  # type: () -> Command
        return self._command

    @property
    def status_code(self):  # type: () -> int
        return self._status_code

    def is_handled(self):  # type: () -> bool
        return self._handled

    def handled(self, handled):  # type: (bool) -> None
        self._handled = handled

    def set_status_code(self, status_code):  # type: (int) -> None
        self._status_code = status_code



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from clikit.api.formatter import Style


class BorderStyle:
    """
    Defines the style of a border.
    """

    _none = None
    _ascii = None
    _solid = None

    def __init__(self):  # type: () -> None
        self.line_ht_char = "-"
        self.line_hc_char = "-"
        self.line_hb_char = "-"

        self.line_vl_char = "|"
        self.line_vc_char = "|"
        self.line_vr_char = "|"

        self.corner_tl_char = "+"
        self.corner_tr_char = "+"
        self.corner_bl_char = "+"
        self.corner_br_char = "+"

        self.crossing_c_char = "+"
        self.crossing_l_char = "+"
        self.crossing_t_char = "+"
        self.crossing_r_char = "+"
        self.crossing_b_char = "+"

        self.style = None  # type: Style

    @classmethod
    def none(cls):  # type: () -> BorderStyle
        if cls._none is None:
            style = cls()
            style.line_ht_char = ""
            style.line_hc_char = ""
            style.line_hb_char = ""

            style.line_vl_char = ""
            style.line_vc_char = " "
            style.line_vr_char = ""

            style.corner_tl_char = ""
            style.corner_tr_char = ""
            style.corner_bl_char = ""
            style.corner_br_char = ""

            style.crossing_c_char = ""
            style.crossing_l_char = ""
            style.crossing_t_char = ""
            style.crossing_r_char = ""
            style.crossing_b_char = ""

            cls._none = style

        return cls._none

    @classmethod
    def ascii(cls):  # type: () -> BorderStyle
        if cls._ascii is None:
            style = cls()

            cls._ascii = style

        return cls._ascii

    @classmethod
    def solid(cls):  # type: () -> BorderStyle
        if cls._solid is None:
            style = cls()
            style.line_ht_char = "─"
            style.line_hc_char = "─"
            style.line_hb_char = "─"

            style.line_vl_char = "│"
            style.line_vc_char = "│"
            style.line_vr_char = "│"

            style.corner_tl_char = "┌"
            style.corner_tr_char = "┐"
            style.corner_bl_char = "└"
            style.corner_br_char = "┘"

            style.crossing_c_char = "┼"
            style.crossing_l_char = "├"
            style.crossing_r_char = "┤"
            style.crossing_t_char = "┬"
            style.crossing_b_char = "┴"

            cls._solid = style

        return cls._solid



from .event import Event


class ConfigEvent(Event):
    """
    Dispatched after the configuration is built.

    Use this event to add custom configuration to the application.
    """

    def __init__(self, config):  # type: (ApplicationConfig) -> None
        self._config = config

    @property
    def config(self):  # type: () -> ApplicationConfig
        return self._config
